import { AlgorithmToLayoutProvider, AllLayoutAlgorithms, LayoutFactory } from '@graphpolaris/shared/lib/graph-layout';
import { schemaGraphology2Reactflow, schemaExpandRelation } from '@graphpolaris/shared/lib/schema/schema-utils';
import {
  useSchemaGraph,
  useSchemaLayout,
  useSchemaSettings,
  useSearchResultSchema,
  useSessionCache,
} from '@graphpolaris/shared/lib/data-access/store';
import { SmartBezierEdge, SmartStepEdge, SmartStraightEdge } from '@tisoap/react-flow-smart-edge';

import { useEffect, useMemo, useRef, useState } from 'react';
import ReactFlow, {
  ControlButton,
  Controls,
  Node,
  Edge,
  ReactFlowProvider,
  useNodesState,
  useEdgesState,
  ReactFlowInstance,
  useReactFlow,
} from 'reactflow';
import CachedIcon from '@mui/icons-material/Cached';
import SettingsIcon from '@mui/icons-material/Settings';

import 'reactflow/dist/style.css';

import styles from './schema.module.scss';

import { ConnectionDragLine, ConnectionLine } from '@graphpolaris/shared/lib/querybuilder/pills';
import { EntityNode } from '../pills/nodes/entity/entity-node';
import { RelationNode } from '../pills/nodes/relation/relation-node';
import NodeEdge from '../pills/edges/node-edge';
import SelfEdge from '../pills/edges/self-edge';
import { useSchemaAPI } from '../../data-access';
import { SchemaDialog } from './schemaDialog';
import { toSchemaGraphology } from '../../data-access/store/schemaSlice';

interface Props {
  content?: string;
  auth?: boolean;
}

const onInit = (reactFlowInstance: ReactFlowInstance) => {
  setTimeout(() => reactFlowInstance.fitView(), 100);
};

const nodeTypes = {
  entity: EntityNode,
  relation: RelationNode,
  // nodeQualityEntityPopup: NodeQualityEntityPopupNode,
  // nodeQualityRelationPopup: NodeQualityRelationPopupNode,
  // attributeAnalyticsPopupMenu: AttributeAnalyticsPopupMenu,

  // entity: EntityRFPill,
  // relation: RelationRFPill,
  // attribute: AttributeRFPill,
};
const edgeTypes = {
  nodeEdge: NodeEdge,
  selfEdge: SelfEdge,
  // connection: ConnectionLine,
  bezier: SmartBezierEdge,
  connection: ConnectionLine,
  straight: SmartStraightEdge,
  step: SmartStepEdge,
};

export const Schema = (props: Props) => {
  const api_schema = useSchemaAPI();
  const session = useSessionCache();
  const settings = useSchemaSettings();

  const searchResults = useSearchResultSchema();

  const [toggleSchemaSettings, setToggleSchemaSettings] = useState(false);
  const [nodes, setNodes, onNodeChanged] = useNodesState([] as Node[]);
  const [edges, setEdges, onEdgeChanged] = useEdgesState([] as Edge[]);
  const [firstUserConnection, setFirstUserConnection] = useState<boolean>(true);
  const [auth, setAuth] = useState(props.auth);
  const settingsIconRef = useRef<SVGSVGElement>(null);
  const dialogRef = useRef<HTMLDivElement>(null);

  // In case the schema is updated
  const schemaGraph = useSchemaGraph();
  const schemaGraphology = useMemo(() => toSchemaGraphology(schemaGraph), [schemaGraph]);
  const layout = useRef<AlgorithmToLayoutProvider<AllLayoutAlgorithms>>();

  function updateLayout() {
    const layoutFactory = new LayoutFactory();
    layout.current = layoutFactory.createLayout(settings.layoutName);
  }

  useEffect(() => {
    updateLayout();
    sessionStorage.setItem('firstUserConnection', firstUserConnection.toString());
  }, []);

  useEffect(() => {
    setAuth(props.auth);
  }, [props.auth]);

  useEffect(() => {
    if (schemaGraphology === undefined || schemaGraphology.order == 0) {
      return;
    }

    updateLayout();
    const expandedSchema = schemaExpandRelation(schemaGraphology);
    layout.current?.layout(expandedSchema);

    const schemaFlow = schemaGraphology2Reactflow(expandedSchema, settings.connectionType);

    // schemaFlow.nodes.forEach((n) => {
    //   n.data.toggleNodeQualityPopup = toggleNodeQualityPopup;
    //   n.data.toggleAttributeAnalyticsPopupMenu =
    //     toggleAttributeAnalyticsPopupMenu;
    // });
    // console.log(edges);

    // console.log(
    //   'schema Layout',
    //   schemaLayout,
    //   'order',
    //   expandedSchema,
    //   schemaFlow
    // );
    setNodes(schemaFlow.nodes);
    setEdges(schemaFlow.edges);
    // console.log(
    //   'update schema useEffect',
    //   schemaGraphology,
    //   schemaGraphology.order,
    //   schemaFlow
    // );
  }, [schemaGraph, settings]);

  useEffect(() => {
    setNodes((nds) =>
      nds.map((node) => ({
        ...node,
        selected: searchResults.includes(node.id) || searchResults.includes(node.data.label),
      }))
    );
  }, [searchResults]);

  useEffect(() => {
    if (dialogRef.current && settingsIconRef.current) {
      dialogRef.current.style.top = `${settingsIconRef.current.getBoundingClientRect().top}px`;
      dialogRef.current.style.left = `${settingsIconRef.current.getBoundingClientRect().left + 30}px`;
    }
  }, [settingsIconRef, dialogRef, toggleSchemaSettings]);

  return (
    <div className="schema-panel w-full h-full">
      <SchemaDialog open={toggleSchemaSettings} onClose={() => setToggleSchemaSettings(false)} ref={dialogRef} />
      <div className="flex justify-between h-[1rem]">
        <h1>Schema</h1>
        <p className="text-slate-400" title={JSON.stringify(session)}>
          Connected to: {session.currentDatabase}
        </p>
      </div>
      {nodes.length === 0 && <p>No Elements</p>}
      <ReactFlowProvider>
        <div className="h-[calc(100%-.8rem)] w-full">
          <ReactFlow
            snapGrid={[10, 10]}
            snapToGrid
            onlyRenderVisibleElements={false}
            nodesDraggable={true}
            nodeTypes={nodeTypes}
            edgeTypes={edgeTypes}
            connectionLineComponent={ConnectionDragLine}
            onNodesChange={onNodeChanged}
            onEdgesChange={onEdgeChanged}
            nodes={nodes}
            edges={edges}
            onInit={onInit}
            proOptions={{ hideAttribution: true }}
          >
            <Controls showInteractive={false} showZoom={false} showFitView={true} className={`${styles.controls} schema-settings`}>
              {/* <ControlButton
              className={styles.exportButton}
              title={'Export graph schema'}
              onClick={(event) => {
                event.stopPropagation();
                // this.setState({
                //   ...this.state,
                //   exportMenuAnchor: event.currentTarget,
                // });
              }}
            >
              <img src={exportIcon} width={21}></img>
            </ControlButton> */}
              <ControlButton
                className={styles.exportButton}
                title={'Refresh graph schema'}
                onClick={(event) => {
                  event.stopPropagation();
                  api_schema.RequestSchema(session.currentDatabase);
                }}
              >
                <CachedIcon />
              </ControlButton>
              <ControlButton
                className={styles.exportButton}
                title={'Open Settings'}
                onClick={(event) => {
                  event.stopPropagation();
                  setToggleSchemaSettings(!toggleSchemaSettings);
                }}
              >
                <SettingsIcon ref={settingsIconRef} />
              </ControlButton>
            </Controls>
          </ReactFlow>
        </div>
      </ReactFlowProvider>
    </div>
  );
};

export default Schema;
